#!/bin/sh

# Copyright (C) 2017-2019 Centrify Corporation. All rights reserved.
#
# install.sh - Centrify DirectAudit install script for CoreOS.
#

CIMS_USR=/opt/centrify
CIMS_INSTALL=/opt/centrify
CIMS_INSTALL_DASH_CDASHMOD="$CIMS_INSTALL/libexec"
export PATH=$CIMS_INSTALL/bin:$CIMS_INSTALL/sbin:/bin:/usr/bin:$PATH

PKG_NAME=CentrifyDA
PKG_NAME_CDC=CentrifyDC
REQUIRED_PKGS="CentrifyDC-openssl centrify-agent"
DEPENDENT_PKGS=
LOG_FILE=/var/log/centrifydc-install.log

PRE_EXT=.pre_cda
TMP_EXT=.tmp
LNK_EXT=.pre_cda_link
LST_EXT=.lst
SYSTEMD_DIR=/etc/systemd/system

TRUE=0
FALSE=1
DO_LOG=$FALSE
DO_FORCE=$FALSE
DO_VERBOSE=$FALSE
DO_SKIP=$FALSE
ACT=
ACTED=
ACTING=
RUN_RESULT=
UPDATE_RESULT=

PROG="`basename $0 .sh`"
PROG_FULL="$0"
COMMAND_LINE="$0 $*"
TGZ=`ls $PKG_NAME-[0-9]*-files.tgz 2> /dev/null | sort -r | head -1`
VERSION="`echo $TGZ | cut -d'-' -f2-3`"
TGZ_FILELIST_NEW="$PKG_NAME-$VERSION$LST_EXT"

TMP_FILE="/tmp/$PROG.sh.$$"
LOG_ERROR="/dev/null"

# Signals need to be caught: 1=SIGHUP 2=SIGINT 3=SIGQUIT 15=SIGTERM
SIGNAL_LIST="1 2 3 15"

GET_PROCESS_PID="PROCESS_PID=\`pgrep -x \$PROCESS\` && test -n \"\$PROCESS_PID\" && ps -o pid= -o cgroup= \$PROCESS_PID | grep -v '/docker/' | grep -v  '/system.slice/docker-' | awk '{ print \$1 }'"

usage()
{
    cat <<EOF

This script installs/uninstalls Centrify DirectAudit service.
Only the superuser can run this script.

Usage:
  $PROG.sh [options...]

where:
  -i            Install
  -e            Uninstall
  -u            Upgrade
  -r            Reinstall
  -f            Force operation (use with -e option)
  -s            Skip package scripts
  -p            Print verbose messages
  -h            Display this help page

EOF
}

log()
{
    if [ "$DO_LOG" = $TRUE ]; then
        echo "$*" >> "$LOG_FILE"
    fi
}

print()
{
    echo "$*"
    log "$@"
}

print2()
{
    echo "$*" >&2
    log "$@"
}

verbose()
{
    if [ "$DO_VERBOSE" = $TRUE ]; then
        print "$@"
    else
        log "$@"
    fi
}

warn()
{
    print2 "WARNING:" "$@"
}

error()
{
    print2 "ERROR:" "$@"
}

run()
{
    verbose "# $*"
    RUN_RESULT=`eval $@ 2>> "$LOG_ERROR"`
    verbose "$RESULT"
}

symlink_to_file()
{
    if [ -h "$1" ]; then
        FILE=`readlink -f "$1"`
        # Backup symlink
        test -e "$1$LNK_EXT" || \
            cp -pfd "$1" "$1$LNK_EXT" 2> /dev/null || true
        # Use mv to make the change atomic
        verbose "Replacing symlink $1 with $FILE ..."
        rm -f "$1$TMP_EXT"
        cp -pf "$FILE" "$1$TMP_EXT"
        mv -f "$1$TMP_EXT" "$1"
    fi
}

install_systemd_service()
{
    if [ -n "$1" ]; then
        verbose "Installing $1 ..."
        cp -pf "$CIMS_INSTALL/etc/systemd/system/$1" \
            "$SYSTEMD_DIR"
    fi
}

is_directory_clean()
{
    if [ -n "$1" ]; then
        verbose "Checking directory $1 ..."
        RUN_RESULT=`find $1/ ! \( -path "$1/" -o -path "$TGZ_FILELIST_INSTALLED" \) 2> /dev/null` || true
        if [ -n "$RUN_RESULT" ]; then
            verbose "File(s) left in $1:"
            for FILE in $RUN_RESULT ; do
                verbose "  $FILE"
            done
            return 1
        fi
    fi
    return 0
}

remove_directory()
{
    if [ -n "$1" ]; then
        is_directory_clean "$1"
        verbose "Removing directory $1 ..."
        rm -rf "$1"
    fi
}

get_version_number()
{
    if [ -n "$1" ]; then
        echo "$1" | sed 's/[^0-9]*//g'
    else
        echo 0
    fi
}

get_current_version()
{
    echo "$TGZ_FILELIST_INSTALLED" | sed -e "s/$LST_EXT$//" | cut -d'-' -f2-3
}

get_package_version()
{
    PKG_FILELIST="`ls -1 $CIMS_INSTALL/$1-[0-9]*$LST_EXT 2> /dev/null`"
    if test -r "$PKG_FILELIST" ; then
        basename "$PKG_FILELIST" | sed -e "s/^$1-//" -e "s/$LST_EXT$//" | cut -d'-' -f1-2
    else
        PKG_FILELIST="`ls -1 $CIMS_INSTALL/etc/$1 2> /dev/null`"
        if test -r "$PKG_FILELIST" ; then
            cat "$PKG_FILELIST"
        fi
    fi
}

get_installed_dependent_packages()
{
    INSTALLED_PKGS=
    for DEPENDENT_PKG in $DEPENDENT_PKGS ; do
        DEPENDENT_VERSION="`get_package_version $DEPENDENT_PKG`"
        if test -n "$DEPENDENT_VERSION" ; then
            # Installed
            INSTALLED_PKGS="$INSTALLED_PKGS $DEPENDENT_PKG"
        fi
    done
    echo $INSTALLED_PKGS
}

are_all_required_packages_installed()
{
    ALL_INSTALLED=$TRUE
    for REQUIRED_PKG in $REQUIRED_PKGS ; do
        REQUIRED_VERSION="`get_package_version $REQUIRED_PKG`"
        if test -z "$REQUIRED_VERSION" ; then
            # Not installed
            verbose "Required package $REQUIRED_PKG is not installed."
            ALL_INSTALLED=$FALSE
        fi
    done
    return $ALL_INSTALLED
}

is_installed()
{
    TGZ_FILELIST_INSTALLED="`ls -1 $CIMS_INSTALL/$PKG_NAME-[0-9]*$LST_EXT 2> /dev/null`"
    test -r "$TGZ_FILELIST_INSTALLED"
}

install_tgz()
{
    print "Start $ACTING $PKG_NAME $VERSION..."
    (
        set -e

        test "$DO_SKIP" = $TRUE || install_tgz_pre

        verbose "Extracting $TGZ ..."
        gunzip -c "$TGZ" | (cd / && tar xfp -) 2>> "$LOG_ERROR"

        test "$DO_SKIP" = $TRUE || install_tgz_post

        tar tf "$TGZ" | tac > "$CIMS_INSTALL/$TGZ_FILELIST_NEW" 2>> "$LOG_ERROR"
        cp -pf $PROG_FULL $CIMS_INSTALL/bin/ 2>> "$LOG_ERROR"
        echo "$CIMS_INSTALL/bin/`basename $PROG_FULL`" >> $CIMS_INSTALL/$TGZ_FILELIST_NEW 2>> "$LOG_ERROR"

        if [ "$ACT" = upgrade -a "$VERSION" != "$CURRENT_VERSION" ]; then

            # Remove files and symlinks that is not in the new package
            verbose "Cleaning old files ..."
            while read -r line ; do
                if ! grep "^$line$" "$CIMS_INSTALL/$TGZ_FILELIST_NEW" > /dev/null ; then
                    if [ -f "/$line" -o -h "/$line" ]; then
                        verbose "  deleting /$line"
                        rm -f "/$line" || true
                    fi
                fi
            done < "$TGZ_FILELIST_INSTALLED" 2>> "$LOG_ERROR"

            # Remove directories that is not in the new package
            verbose "Cleaning old directories ..."
            DIR_REMOVED=$TRUE
            while test "$DIR_REMOVED" = $TRUE ; do
                DIR_REMOVED=$FALSE
                while read -r line ; do
                    if ! grep "^$line$" "$CIMS_INSTALL/$TGZ_FILELIST_NEW" > /dev/null ; then
                        if [ -d "/$line" -a "/$line" != "/./" ]; then
                            # Directory may belong to system,
                            # do not show error
                            verbose "  deleting /$line"
                            rmdir "/$line" 2> /dev/null && DIR_REMOVED=$TRUE || true
                        fi
                    fi
                done < "$TGZ_FILELIST_INSTALLED" 2>> "$LOG_ERROR"
            done

            rm "$TGZ_FILELIST_INSTALLED" 2>> "$LOG_ERROR" || true
        fi
    )

    RC=$?
    if [ "$ACT" = install -a $RC -ne 0 ]; then
        rm -f "$CIMS_INSTALL/$TGZ_FILELIST_NEW"
    fi

    return $RC
}

uninstall_tgz()
{
    print "Start $ACTING $PKG_NAME $CURRENT_VERSION ..."
    (
        set -e

        if [ ! -r "$TGZ_FILELIST_INSTALLED" ]; then
            error "Cannot read $TGZ_FILELIST_INSTALLED"
            false
        fi

        test "$DO_SKIP" = $TRUE || uninstall_tgz_pre

        # Remove files and symlinks
        verbose "Removing package files ..."
        while read -r line ; do
            if [ -f "/$line" -o -h "/$line" ]; then
                rm -f "/$line" || true
            fi
        done < "$TGZ_FILELIST_INSTALLED" 2>> "$LOG_ERROR"

        # Remove directories
        verbose "Removing package directories ..."
        DIR_REMOVED=$TRUE
        while test "$DIR_REMOVED" = $TRUE ; do
            DIR_REMOVED=$FALSE
            while read -r line ; do
                if [ -d "/$line" -a "/$line" != "/./" ]; then
                    # Directory may belong to system,
                    # do not show error
                    rmdir "/$line" 2> /dev/null && DIR_REMOVED=$TRUE || true
                fi
            done < "$TGZ_FILELIST_INSTALLED" 2>> "$LOG_ERROR"
        done

        test "$DO_SKIP" = $TRUE || uninstall_tgz_post

        rm -f "$TGZ_FILELIST_INSTALLED"
    )
    
    if [ "$DO_FORCE" = $TRUE ] && is_installed ; then
        print "Force $ACTING ..."
        rm -f "$TGZ_FILELIST_INSTALLED"
    fi

    test ! -f "$TGZ_FILELIST_INSTALLED"
}

install_tgz_pre()
{
    verbose "Running pre-install script ..."

    # Check if /etc/logrotate.d/centrifydc is a symlink
    LOGROTATE_CONF="/etc/logrotate.d/centrifydc"
    if [ -h "${LOGROTATE_CONF}" ]; then
        echo "ERROR: The file ${LOGROTATE_CONF} is a symbolic link which is not expected." \
            "Cannot proceed with upgrade/installation." \
            "Please contact Centrify Technical Support." | tee -a "$LOG_ERROR"
        exit 1
    fi

    if [ "$ACT" = upgrade -o "$ACT" = reinstall ]; then
        verbose "Pre-install for $ACTING to $VERSION ..."

        CDA_CUR_VER=$CURRENT_VERSION
        log "CDA_CUR_VER=$CDA_CUR_VER"

        if [ `$CIMS_USR/bin/dainfo -s > /dev/null 2>&1; echo $?` -ne 0 ]; then
            verbose "Centrify DirectAudit auditing is currently enabled (symlink hooked)."
            IS_AUDITING=$TRUE
        else
            IS_AUDITING=$FALSE
        fi

        if [ ${IS_AUDITING} -eq $FALSE ]; then
            $CIMS_USR/sbin/dacontrol --cdashContWithoutDad >> "$LOG_ERROR" 2>&1 || true

            for tries in 1 2 3 4 5; do
                dadPid=`PROCESS=dad && eval $GET_PROCESS_PID` || true
                if [ ! -z "$dadPid" ]; then
                    case "$tries" in
                    1)
                        verbose "Stopping dad by centrifyda stop ..."
                        $CIMS_INSTALL/bin/centrifyda stop  >> "$LOG_ERROR" 2>&1
                        ;;
                    2)
                        verbose "Stopping dad by dastop ..."
                        $CIMS_USR/sbin/dastop >> "$LOG_ERROR" 2>&1 || true
                        ;;
                    3 | 4)
                        verbose "Stopping dad by kill ..."
                        kill $dadPid >> "$LOG_ERROR" 2>&1 || true
                        ;;
                    5)
                        verbose "Stopping dad by kill ..."
                        kill -9 $dadPid >> "$LOG_ERROR" 2>&1 || true
                        ;;
                    esac
                    sleep 5
                else
                    verbose "DirectAudit daemon status: not running ..."
                    break
                fi
            done

            # Save configuration files
            verbose "Saving configuration files for $ACT ..."
            mkdir -p /etc/centrifyda/custom >> "$LOG_ERROR" 2>&1 || true
            cp -pf /etc/centrifyda/*.* \
                /etc/centrifyda/custom >> "$LOG_ERROR" 2>&1 || true
        else
            print "Upgrading from ${CDA_CUR_VER} requires disabling DirectAudit so that the new"
            print "DirectAudit mechanism for hooking shells can be installed."
            print "Please run 'dacontrol -d -a' to disable auditing, and then restart upgrading."
            exit 1
        fi
    
        # Environment unsuitable to install should have exited. Remove $CIMS_INSTALL_DASH_CDASHMOD/centrifyda/
        if [ -d $CIMS_INSTALL_DASH_CDASHMOD/centrifyda ]; then
            rm -Rf $CIMS_INSTALL_DASH_CDASHMOD/centrifyda || true
        fi
    fi
}

install_tgz_post()
{
    verbose "Running post-install script ..."
    # Install centrifyda systemd service
    verbose "Installing Centrify DirectAudit systemd service ..."
    umask 0022
    mkdir -p "$SYSTEMD_DIR" >> "$LOG_ERROR" 2>&1 || true
    install_systemd_service centrifyda.service >> "$LOG_ERROR" 2>&1 || true
    systemctl daemon-reload >> "$LOG_ERROR" 2>&1 || true

    if [ "$ACT" = upgrade -o "$ACT" = reinstall ]; then
        # Save new files
        mkdir -p /etc/centrifyda/new >> "$LOG_ERROR" 2>&1 || true
        cp -pf /etc/centrifyda/*.* /etc/centrifyda/new/. >> "$LOG_ERROR" 2>&1 || true

        # Restore customized files
        cp -pf /etc/centrifyda/custom/*.* /etc/centrifyda/. >> "$LOG_ERROR" 2>&1 || true

        # Output directory
        rm -rf /etc/centrifyda/merge >> "$LOG_ERROR" 2>&1 || true
        mkdir -p /etc/centrifyda/merge >> "$LOG_ERROR" 2>&1 || true

        # Upgrade configuration files
        verbose "Upgrading configuration files ..."

        RC=0
        $CIMS_INSTALL/bin/upgradeconf -r -c \
            /etc/centrifyda/new/upgradeconf.conf \
            /etc/centrifyda/custom/centrifyda.conf \
            /etc/centrifyda/new/centrifyda.conf \
            /etc/centrifyda/merge/centrifyda.conf >> "$LOG_ERROR" 2>&1 || RC=$?

        # If upgrade successfully, update the config files
        if [ $RC -eq 0 ] ; then
            verbose "Replacing configuration files ..."
            for FILE in /etc/centrifyda/merge/*.* ; do
                if [ -f "$FILE" -a -s "$FILE" ]; then
                    verbose "  moving $FILE to /etc/centrifyda/"
                    mv -f "$FILE" /etc/centrifyda/. >> "$LOG_ERROR" 2>&1 || true
                fi
            done
            rm -rf /etc/centrifyda/merge >> "$LOG_ERROR" 2>&1 || true
        else
            verbose "Failed to merge configuration files."
        fi

        # Save installation ID file
        if [ -f /var/centrifydc/kset.da.installation.id ]; then
            mv -f /var/centrifydc/kset.da.installation.id /var/centrifyda/kset.da.installation.id >> "$LOG_ERROR" 2>&1 || true
        fi

        $CIMS_USR/sbin/dacontrol --cdashContWithoutDad >> "$LOG_ERROR" 2>&1 || true
        verbose "Restarting DirectAudit daemon ..."
        for tries in 1 2 3 4 5; do
            dadPid=`PROCESS=dad && eval $GET_PROCESS_PID` || true
            if [ ! -z "$dadPid" ]; then
                case "$tries" in
                1)
                    verbose "Stopping dad by centrifyda stop ..."
                    $CIMS_INSTALL/bin/centrifyda stop >> "$LOG_ERROR" 2>&1 || true
                    ;;
                2)
                    verbose "Stopping dad by dastop ..."
                    $CIMS_USR/sbin/dastop >> "$LOG_ERROR" 2>&1 || true
                    ;;
                3 | 4)
                    verbose "Stopping dad by kill ..."
                    kill $dadPid >> "$LOG_ERROR" 2>&1 || true
                    ;;
                5)
                    verbose "Stopping dad by kill ..."
                    kill -9 $dadPid >> "$LOG_ERROR" 2>&1 || true
                    ;;
                esac
                sleep 5
            else
                verbose "DirectAudit daemon status: not running ..."
                break
            fi
        done 
    fi

    # If ".$CIMS_INSTALL/system/lib/" is found in neither CDC core nor add-on packages' filelist, 
    # insert it into CDC core package's filelist before the line of ".$CIMS_INSTALL/system/",
    # then directory "$CIMS_INSTALL/system/lib/" can be removed when CDC is being uninstalled.
    CDC_PKG_FILELIST="`ls -1 $CIMS_INSTALL/$PKG_NAME_CDC-*$LST_EXT 2> /dev/null`"
    FOUND=0
    for CDC_PKG_TEMP in $CDC_PKG_FILELIST
    do
        if grep ".$CIMS_INSTALL/system/lib/$" $CDC_PKG_TEMP > /dev/null ; then
            FOUND=1
            break
        fi
    done
    
    if [ $FOUND -eq 0 ]; then
        verbose "Update CentrifyDC package info to fix a package issue."
        FILELIST_CDC="`ls -1 $CIMS_INSTALL/$PKG_NAME_CDC-[0-9]*$LST_EXT 2> /dev/null`"
        FILELIST_CDC_TMP="${FILELIST_CDC}.tmp"
        sed "\:$CIMS_INSTALL/system/$:i.$CIMS_INSTALL/system/lib/" $FILELIST_CDC > $FILELIST_CDC_TMP 2>> "$LOG_ERROR" && \
            mv -f $FILELIST_CDC_TMP $FILELIST_CDC 2>> "$LOG_ERROR" || true
    fi
    
    # Enable centrifyda systemd service
    verbose "Enabling Centrify DirectAudit service ..."
    systemctl enable centrifyda.service 2>> "$LOG_ERROR" || true

    # Create cdax parent directory for each shell
    SHELL_LIST=`cat /etc/shells | grep -v -e ".*/cdax/.*" -e "^\s*#" | tr "\n" " "`
    INVALID_SHELL_LIST="nologin passwd true false centrifyda"
    OLD_IFS=$IFS
    IFS=" "
    for SHELL_TEMP in $SHELL_LIST
    do
        SHELL_DIRECTORY=`dirname $SHELL_TEMP`
        SHELL_NAME=`basename $SHELL_TEMP`
        INVALID=0

        # Check if the shell is a invalid shell
        for INVALID_SHELL in $INVALID_SHELL_LIST 
        do
            if [ $SHELL_NAME = $INVALID_SHELL ]; then
                INVALID=1
                break
            fi
        done

        # If the shell is invalid shell, not create cdax parent directory for it.
        if [ $INVALID -ne 0 ]; then
            continue
        fi

        if [ ! -d "$CIMS_INSTALL$SHELL_DIRECTORY" ]; then
            mkdir -p  "$CIMS_INSTALL$SHELL_DIRECTORY" >> "$LOG_ERROR" 2>&1 || true
        fi
    done
    IFS=$OLD_IFS

    # Backup symlink file /etc/shells and create a new regular one
    symlink_to_file /etc/shells >> "$LOG_ERROR" 2>&1 || true

    verbose "Starting DirectAudit daemon ..."
    $CIMS_INSTALL/bin/centrifyda start >> "$LOG_ERROR" 2>&1 || true

    $CIMS_USR/sbin/dacontrol --cdashHangWithoutDad >> "$LOG_ERROR" 2>&1 || true

    chmod 444 $CIMS_INSTALL/etc/centrifyda_suite_version >> "$LOG_ERROR" 2>&1 || true
    chmod 444 $CIMS_INSTALL/etc/centrifyda_version >> "$LOG_ERROR" 2>&1 || true

    # If it's doing a upgrade and NSS is already enabled, run `dacontrol -e' to make /path/to/shell/cdax/shellname is the latest.
    if [ "$ACT" = upgrade -o "$ACT" = reinstall ] && [ ! -f /var/centrifyda/NSSPurposelyInactivated ]; then
        $CIMS_USR/sbin/dacontrol -e >> "$LOG_ERROR" 2>&1 || true
    fi

    $CIMS_INSTALL/sbin/daflush >> "$LOG_ERROR" 2>&1 || true

    # Reload CentrifyDC to load the newly installed DA plugin
    verbose "Reload DirectControl ..."
    RC=0
    $CIMS_INSTALL/sbin/adreload -c -m > /dev/null 2>&1 || RC=$?
    verbose "adreload returned $RC"

    /sbin/ldconfig
}

uninstall_tgz_pre()
{
    verbose "Running pre-uninstall script ..."
    # Turn dadebug off
    if dadebug | grep off >> "$LOG_ERROR" 2>&1 ; then
        :
    else
        verbose "Disabling dadebug ..."
        
        if ! dadebug off >> "$LOG_ERROR" 2>&1 ; then
            warn "Cannot disable dadebug"
        fi
    fi

    $CIMS_INSTALL/sbin/dacontrol -d >> "$LOG_ERROR" 2>&1 || true
    $CIMS_INSTALL/sbin/dacontrol -d -a >> "$LOG_ERROR" 2>&1 || true
    $CIMS_INSTALL/sbin/dacontrol --cdashContWithoutDad >> "$LOG_ERROR" 2>&1 || true
    $CIMS_INSTALL/bin/centrifyda stop >> "$LOG_ERROR" 2>&1 || true

    # Disable centrifyda systemd service
    systemctl disable centrifyda.service >> "$LOG_ERROR" 2>&1 || true
}

uninstall_tgz_post()
{
    verbose "Running post-uninstall script ..."
    # Reload CentrifyDC to unload the DA plugin
    verbose "Reload DirectControl ..."
    RC=0
    $CIMS_INSTALL/sbin/adreload -c -m > /dev/null 2>&1 || RC=$?
    verbose "adreload returned $RC"

    # Remove files which were created during running centrifyda
    remove_directory /var/centrifyda > /dev/null || true
    remove_directory /etc/centrifyda > /dev/null || true
    rm -f $CIMS_INSTALL/bin/da.emergency.shell /etc/centrifyda/da.emergency.shell $CIMS_INSTALL/libexec/da.emergency.shell || true

    # Remove centrifyda systemd service
    rm -f "$SYSTEMD_DIR/centrifyda.service" || true

    # Try to remove directory "$CIMS_INSTALL$CIMS_INSTALL_DASH_CDASHMOD", since 
    # $CIMS_INSTALL$CIMS_INSTALL_DASH_CDASHMOD might be created by mistake (CS-44945) 
    # while upgrading from Suite 2017.3 M2 to Suite 2017.3 GA
    rm -rf $CIMS_INSTALL$CIMS_INSTALL_DASH_CDASHMOD
    
    # NextGen DA: need to clean up the following dirs if CDA is installed with CAgent instead of CDC
    if [ ! -f $CIMS_USR/sbin/adclient ]; then
        DIR_LIST=("mappers/machine" "mappers" "perl" "scripts")
        for DIR in ${DIR_LIST[@]}
        do
            rmdir $CIMS_INSTALL/$DIR >> "$LOG_ERROR" 2>&1 || true
        done
    fi    
    
    # Remove .*/cdax/.* shells and its parent directory created by us during installing
    SHELL_LIST=`cat /etc/shells | tr "\n" " "`
    OLD_IFS=$IFS
    IFS=" "
    for SHELL_TEMP in $SHELL_LIST
    do
        SHELL_DIRECTORY=`dirname $SHELL_TEMP`
        if [ -d "$SHELL_DIRECTORY" ]; then
            if echo "$SHELL_DIRECTORY" | grep -q -e "^$CIMS_INSTALL.*/cdax$"; then
                rm -rf "$SHELL_DIRECTORY" || true
                TOP_DIRECTORY=`dirname $SHELL_DIRECTORY`
            else
                TOP_DIRECTORY=$CIMS_INSTALL$SHELL_DIRECTORY
            fi            
            while [ -d $TOP_DIRECTORY -a $CIMS_INSTALL != $TOP_DIRECTORY ]
            do
                rmdir "$TOP_DIRECTORY" 2> /dev/null || break
                TOP_DIRECTORY=`dirname $TOP_DIRECTORY`
            done
        fi
    done
    IFS=$OLD_IFS

    # Remove shells in /etc/shells added by DA
    if [ -f "/etc/shells" ]; then
        cp -pf /etc/shells /etc/shells.new || true
        cat /etc/shells | grep -v "$CIMS_INSTALL/libexec/centrifyda" | grep -v -e ".*/cdax/.*" > /etc/shells.new || true
        mv /etc/shells.new /etc/shells || true
        /sbin/ldconfig || true
    fi
}

clean_up()
{
    rm -f "$TMP_FILE"
}

clean_exit()
{
    clean_up
    exit $1
}

signal_handler()
{
    trap "" ${SIGNAL_LIST}
    error "Signal received."
    clean_up
    exit 1
}


# Parse arguments
while getopts :ieurfsph OPT; do
    case $OPT in
        i)
            ACT=install
            ;;
        e)
            ACT=uninstall
            ;;
        u)
            ACT=upgrade
            ;;
        r)
            ACT=reinstall
            ;;
        f)
            DO_FORCE=$TRUE
            ;;
        s)
            DO_SKIP=$TRUE
            ;;
        p)
            DO_VERBOSE=$TRUE
            ;;
        h)
            usage
            exit 0
            ;;
        \?)
            error "Invalid option '-$OPTARG'." >&2
            exit 1
            ;;
        :)
            error "'-$OPTARG' requires an argument." >&2
            exit 1
            ;;
    esac
done
shift $((OPTIND-1))

if [ -z "$ACT" ]; then
    error "No operation is specified."
    usage
    exit 1
fi

case "$ACTION" in
    install|upgrade|reinstall|uninstall)
        if [ "`id -u`" != 0 ]; then
            echo "This script must be run by root." >&2
            exit 1
        fi
        ;;
esac

# Write a marker to the log file
DO_LOG=$TRUE
LOG_ERROR=$LOG_FILE
log ""
log "--- $PROG.sh ---:" `date`
log "Command line: $COMMAND_LINE"

trap signal_handler ${SIGNAL_LIST}
touch "$TMP_FILE" >> "$LOG_ERROR" 2>&1 || true

# Preprocess for install, upgrade and reinstall
case "$ACT" in
    install|upgrade|reinstall)
        if [ -f "$TGZ" ]; then
            if [ "$ACT" = upgrade -o "$ACT" = reinstall ] && ! is_installed ; then
                ACT=install
            fi
            if [ "$ACT" = install ]; then
                ACTED=installed
                ACTING=installing
                if ! is_installed ; then
                    if ! are_all_required_packages_installed ; then
                        error "Please install all required packages first!"
                        clean_exit 1
                    fi
                else
                    error "$PKG_NAME is already installed."
                    clean_exit 1
                fi
            else
                VERSION_NUM=`get_version_number "$VERSION"`
                CURRENT_VERSION=`get_current_version`
                CURRENT_VERSION_NUM=`get_version_number "$CURRENT_VERSION"`
                if [ -n "$CURRENT_VERSION" ]; then
                    if [ "$ACT" = upgrade ]; then
                        ACTED=upgraded
                        ACTING=upgrading
                        if [ $VERSION_NUM -lt $CURRENT_VERSION_NUM ]; then
                            error "$PKG_NAME installed version $CURRENT_VERSION is newer than $VERSION."
                            clean_exit 1
                        fi
                    else
                        ACTED=reinstalled
                        ACTING=reinstalling
                        if [ "$CURRENT_VERSION" != "$VERSION" ]; then
                            error "$PKG_NAME installed version is not $VERSION."
                            clean_exit 1
                        fi
                    fi
                else
                    error "$PKG_NAME version cannot be determined."
                    clean_exit 1
                fi
            fi
        else
            error "$PKG_NAME tarball not found."
            clean_exit 1
        fi
        ;;
esac

case "$ACT" in
    install|upgrade|reinstall)
        install_tgz
        RC=$?
        if [ $RC -eq 0 ]; then
            print "$PKG_NAME is successfully $ACTED."
        else
            error "Cannot $ACT $PKG_NAME."
            clean_exit 1
        fi
        ;;
    uninstall)
        ACTED=uninstalled
        ACTING=uninstalling
        if is_installed ; then
            CURRENT_VERSION=`get_current_version`
            CURRENT_VERSION_NUM=`get_version_number "$CURRENT_VERSION"`
            DEPENDENT_PKGS=`get_installed_dependent_packages`
            if [ -n "$DEPENDENT_PKGS" ]; then
                error "Cannot $ACT because ${DEPENDENT_PKGS} are currently installed."
                clean_exit 1
            fi

            uninstall_tgz
            RC=$?
            if [ $RC -eq 0 ]; then
                print "$PKG_NAME is successfully $ACTED."
            else
                error "Cannot $ACT $PKG_NAME." >&2
                clean_exit 1
            fi
        else
            error "$PKG_NAME is not installed." >&2
            clean_exit 1
        fi
        ;;
    *)
        error "Unknown operation." >&2
        clean_exit 1
esac

clean_up
exit 0
